package de.tudresden.inf.st.mquat.solving.clustering;

import de.tudresden.inf.st.mquat.jastadd.model.Resource;
import de.tudresden.inf.st.mquat.jastadd.model.ResourceType;
import de.tudresden.inf.st.mquat.jastadd.model.Root;
import de.tudresden.inf.st.mquat.utils.EvalutationException;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import weka.clusterers.SimpleKMeans;
import weka.core.Attribute;
import weka.core.Instances;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

/***
 * Wrapper for WEKA SimpleKMeans. Clustering based on:
 * CPU frequency
 * RAM total
 * Disk total
 * Network throughput
 *
 * @author Dmytro Pukhkaiev
 * @version 0.1
 */
public class KMeans {
  private static final Logger logger = LogManager.getLogger(KMeans.class);
  // dataset
  private Instances resources;
  // the model
  private Root model;
  // list of resources
  private ArrayList<HardwareComponent> hardwareComponentList = new ArrayList<>();

  /***
   * Prepare clustering algorithm, extract WEKA instances from Jastadd model
   * @param model Jastadd model
   */
  public KMeans(Root model) {
    this.model = model;
    // create attributes for instances
    Attribute cpu = new Attribute("cpu");
    Attribute ram = new Attribute("ram");
    Attribute disk = new Attribute("disk");
    Attribute network = new Attribute("network");
    ArrayList<Attribute> attributeList = new ArrayList<>();

    // we expect one top-level resource type and consider at most four sub-resources of the following names
    for (ResourceType resourceType : model.getHardwareModel().getResourceType(0).getSubTypeList()) {
      switch (resourceType.name().toLowerCase()) {
        case "cpu":
          attributeList.add(cpu);
          break;
        case "ram":
          attributeList.add(ram);
          break;
        case "disk":
          attributeList.add(disk);
          break;
        case "network":
          attributeList.add(network);
      }
    }

    // create a dataset
    resources = new Instances("resources", attributeList, 0);

    //instantiate items in the dataset
    for (Resource resource : model.getHardwareModel().getResourceList()) {
      if (resource.getType().getName().getName().equals("ComputeNode")) {
        HardwareComponent hardwareComponent = new HardwareComponent(resource, resource.getNumSubResource());
        hardwareComponent.getInstance().setDataset(resources);

        for (Resource sub : resource.getSubResourceList()) {
          String typeName = sub.getType().getName().getName();
          switch (typeName) {
            case "CPU":
              hardwareComponent.getValue().put("CPU", sub.getCurrentValueByPropertyName("frequency"));
              //hardwareComponent.getInstance().setValue(cpu, sub.getCurrentValueByPropertyName("frequency"));
              break;
            case "RAM":
              hardwareComponent.getValue().put("RAM", sub.getCurrentValueByPropertyName("total"));
              //hardwareComponent.getInstance().setValue(ram, sub.getCurrentValueByPropertyName("total"));
              break;
            case "DISK":
              hardwareComponent.getValue().put("DISK", sub.getCurrentValueByPropertyName("total"));
              //hardwareComponent.getInstance().setValue(disk, sub.getCurrentValueByPropertyName("total"));
              break;
            case "NETWORK":
              hardwareComponent.getValue().put("NETWORK", sub.getCurrentValueByPropertyName("throughput"));
              //hardwareComponent.getInstance().setValue(network, sub.getCurrentValueByPropertyName("throughput"));
              break;
            default:
              throw new RuntimeException("Unable to transform unknown resource of type" + typeName);
          }
        }

        hardwareComponentList.add(hardwareComponent);
      }
    }

    Map<String, Double> minValue = new HashMap<>();
    Map<String, Double> maxValue = new HashMap<>();

    for (HardwareComponent hwc : hardwareComponentList) {
      for (String key : hwc.getValue().keySet()) {
        if (!minValue.containsKey(key))
          minValue.put(key, hwc.getValue().get(key));
        else if (minValue.get(key) > hwc.getValue().get(key))
          minValue.put(key, hwc.getValue().get(key));

        if (!maxValue.containsKey(key))
          maxValue.put(key, hwc.getValue().get(key));
        else if (maxValue.get(key) < hwc.getValue().get(key))
          maxValue.put(key, hwc.getValue().get(key));
      }
    }
    LinearNormalizer linearNormalizer = new LinearNormalizer(minValue, maxValue);

    for (HardwareComponent hwc : hardwareComponentList) {
      hwc.setNormalizedValue(linearNormalizer.normalize(hwc));
      for (String key : hwc.getNormalizedValue().keySet()) {
        switch (key) {
          case "CPU":
            hwc.getInstance().setValue(cpu, hwc.getNormalizedValue().get(key));

            break;
          case "RAM":
            hwc.getInstance().setValue(ram, hwc.getNormalizedValue().get(key));
            break;
          case "DISK":
            hwc.getInstance().setValue(disk, hwc.getNormalizedValue().get(key));
            break;
          case "NETWORK":
            hwc.getInstance().setValue(network, hwc.getNormalizedValue().get(key));

            break;
          default:
            throw new RuntimeException("Unable to transform unknown resource of type" + key);
        }
      }
      resources.add(hwc.getInstance());
    }
  }

  public Map<Integer, ArrayList<Resource>> buildClusters(int numClusters) {

    if (numClusters < 1) {
      logger.error("Unable to cluster the model in less than one cluster (numClusters was " + numClusters + ")");
      throw new EvalutationException();
    }

    logger.info("Clustering " + model.getHardwareModel().getNumResource() + " resources in " + numClusters + " clusters");

    SimpleKMeans algorithm = new SimpleKMeans();
    Map<Integer, ArrayList<Resource>> Clusters = new HashMap<>();

    for (int i = 0; i < numClusters; i++)
      Clusters.put(i, new ArrayList<>());
    try {
      algorithm.setNumClusters(numClusters);
      algorithm.setPreserveInstancesOrder(true);
      algorithm.buildClusterer(resources);

      int[] assignments = algorithm.getAssignments();

      for (int i = 0; i < assignments.length; i++) {
        Clusters.get(assignments[i]).add(hardwareComponentList.get(i).getResource());
      }

    } catch (Exception e) {
      e.printStackTrace();
    }

    int nonEmpty = 0;
    for (int c = 0; c < numClusters; c++) {
      if (Clusters.get(c).size() > 0) {
        nonEmpty++;
      }
    }

    logger.info("Clustering completed, created {} non-empty clusters.", nonEmpty);

    return Clusters;
  }
}
