package de.tudresden.inf.st.mquat.solving;

import de.tudresden.inf.st.mquat.jastadd.model.*;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.junit.Assert;
import org.junit.BeforeClass;
import org.junit.Rule;
import org.junit.rules.ErrorCollector;

import java.util.Iterator;

import static org.hamcrest.core.IsEqual.equalTo;
import static org.junit.Assert.assertEquals;

public abstract class HandwrittenTestSuite {
  private static Logger logger;

  @Rule
  public ErrorCollector collector = new ErrorCollector();

  @BeforeClass
  public static void setupClass() {
    logger = LogManager.getLogger(HandwrittenTestSuite.class);
  }

  protected Assignment assertAssignment(Solution solution, String requestName, String implName, String resourceNamePrefix) {
    Iterator<Assignment> assignmentIterator = solution.assignmentIterator();
    while (assignmentIterator.hasNext()) {
      Assignment assignment = assignmentIterator.next();
      Request request = assignment.getRequest();
      logger.debug("checking " + request.name() + " " + assignment.getImplementation().name() + " " + assignment.getResource().name());
      if (request.name().equals(requestName) && assignment.getImplementation().name().equals(implName)) {
        Assert.assertTrue("The resource starts with the given prefix.", assignment.getMapping().asResourceMapping().getResource().name().startsWith(resourceNamePrefix));
        return assignment;
      }
    }
    Assert.fail("Expected assignment not found.");
    return null;
  }

  protected Assignment assertAssignment(Tuple<Root, Solution> modelAndSolution,
                                        int request, String impl, String resource) {
    Root model = modelAndSolution.getFirstElement();
    Assignment expectedAssignment = new Assignment();
    expectedAssignment.setRequest(model.getRequest(request));
    expectedAssignment.setImplementation(model.resolveImplementation(impl).get());
    expectedAssignment.setMapping(new ResourceMapping(expectedAssignment.getImplementation().getResourceRequirement().getInstance(0), model.resolveResource(resource).get(), new List<>()));
    // check if assignment matches (partly) one listed in the solution
    Iterator<Assignment> assignmentIterator = modelAndSolution.getSecondElement().assignmentIterator();
    while (assignmentIterator.hasNext()) {
      Assignment actualAssignment = assignmentIterator.next();
      if (matches(actualAssignment, expectedAssignment)) {
        Assert.assertNotNull(actualAssignment);
        return actualAssignment;
      }
    }
    Assert.fail(String.format("Did not find match of assignment: %s on %s for request%s",
        impl, resource, request));
    throw new AssertionError();
  }

  protected void assertComponentRequirement(Assignment requiringAssignment,
                                            String instanceName, Assignment expectedProvidingAssignment) {
    Instance instance = requiringAssignment.getImplementation().resolveInstance(instanceName).get();
    Assignment actualProvidingAssignment = requiringAssignment.mappedAssignment(instance);
    assertEquals(String.format("Not matching assignment for %s", instanceName),
        expectedProvidingAssignment, actualProvidingAssignment);
  }

  /**
   * Check if request, configuration and resource of the given assignments are equal
   *
   * @param actualAssignment   assignment in the computed solution
   * @param expectedAssignment expected assignment defined in the test case
   * @return <code>true</code> if both match, <code>false</code> otherwise
   */
  protected boolean matches(Assignment actualAssignment, Assignment expectedAssignment) {
    return actualAssignment.getRequest().equals(expectedAssignment.getRequest()) &&
        actualAssignment.getImplementation().equals(expectedAssignment.getImplementation()) &&
        actualAssignment.getResource().equals(expectedAssignment.getResource());
  }

  protected void assertValidSolution(Tuple<Root, Solution> modelAndSolution) {
//    Assert.assertTrue("Solution is not valid", modelAndSolution.getSecondElement().isValid());
    if (modelAndSolution.getSecondElement().getNumAssignment() == 0) {
      logger.warn("Solution is empty");
    }
    collector.checkThat("Solution is not valid", true,
        equalTo(modelAndSolution.getSecondElement().isValid()));

    logger.info("completeness: " + modelAndSolution.getSecondElement().evaluateSoftwareCompleteness());
    logger.info("sw-validity: " + modelAndSolution.getSecondElement().evaluateSoftwareValidity());
    logger.info("validity: " + modelAndSolution.getSecondElement().evaluateValidity());
    logger.info("objective: " + modelAndSolution.getSecondElement().computeObjective());
  }

}
