package de.tudresden.inf.st.mquat.solving;

import beaver.Parser;
import de.tudresden.inf.st.mquat.jastadd.model.*;
import de.tudresden.inf.st.mquat.utils.ParserUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.junit.Test;

import java.io.IOException;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

public abstract class OptimalHandwrittenTestSuite extends HandwrittenTestSuite {

  private static Logger logger = LogManager.getLogger(OptimalHandwrittenTestSuite.class);

  /**
   * Create and return the solver to use in this test suite
   * @return a solver instance
   */
  protected abstract Solver getSolver();

  protected Tuple<Root, Solution> loadAndSolve(String filename) throws IOException, Parser.Exception, SolvingException {
    Root model = ParserUtils.load(filename, HandwrittenTestSuite.class);
    Solution solution = getSolver().solve(model);
    if (logger.isDebugEnabled()) {
      MquatString out = solution.print(new MquatWriteSettings(" "));
      logger.debug("Solution:\n{}", out);
    }
    return new Tuple<>(model, solution);
  }

  @Test
  public void test_01() throws IOException, Parser.Exception, SolvingException {
    Tuple<Root, Solution> modelAndSolution = loadAndSolve("test_01.txt");
    assertValidSolution(modelAndSolution);
    Assignment config_0i0m0 = assertAssignment(modelAndSolution, 0, "config_0i0m0", "r0");
    Assignment config_1i0m0 = assertAssignment(modelAndSolution, 0, "config_1i0m0", "r1");
    assertComponentRequirement(config_0i0m0,"other", config_1i0m0);
  }

  @Test
  public void test_02() throws IOException, Parser.Exception, SolvingException {
    Tuple<Root, Solution> modelAndSolution = loadAndSolve("test_02.txt");
    assertValidSolution(modelAndSolution);
    Assignment config_0i0m0 = assertAssignment(modelAndSolution, 0, "config_0i0m0", "r0");
    Assignment config_1i0m0 = assertAssignment(modelAndSolution, 0, "config_1i0m0", "r1");
    assertComponentRequirement(config_0i0m0,"other", config_1i0m0);
  }

  @Test
  public void test_03() throws IOException, Parser.Exception, SolvingException {
    Tuple<Root, Solution> modelAndSolution = loadAndSolve("test_03.txt");
    assertValidSolution(modelAndSolution);
    Assignment r0config_0i0m0 = assertAssignment(modelAndSolution, 0, "config_0i0m0", "r0");
    Assignment r0config_1i0m0 = assertAssignment(modelAndSolution, 0, "config_1i0m0", "r1");
    assertAssignment(modelAndSolution, 1, "config_1i0m0", "r2");
    assertComponentRequirement(r0config_0i0m0,"other", r0config_1i0m0);
  }

  @Test
  public void test_04() throws IOException, Parser.Exception, SolvingException {
    Tuple<Root, Solution> modelAndSolution = loadAndSolve("test_04.txt");
    assertValidSolution(modelAndSolution);
    Assignment config_0i0m0 = assertAssignment(modelAndSolution, 0, "config_0i0m0", "r0");
    Assignment config_1i0m0 = assertAssignment(modelAndSolution, 0, "config_1i0m0", "r1");
    Assignment config_2i0m0 = assertAssignment(modelAndSolution, 0, "config_2i0m0", "r2");
    assertComponentRequirement(config_0i0m0,"alpha", config_1i0m0);
    assertComponentRequirement(config_0i0m0,"beta", config_2i0m0);
  }

  @Test
  public void test_05() throws IOException, Parser.Exception, SolvingException {
    Tuple<Root, Solution> modelAndSolution = loadAndSolve("test_05.txt");
    assertValidSolution(modelAndSolution);
    assertTrue(modelAndSolution.getSecondElement().isValid());
    Assignment configA = assertAssignment(modelAndSolution, 0, "configA0", "r0");
    Assignment configB = assertAssignment(modelAndSolution, 0, "configB0", "r1");
    Assignment configC = assertAssignment(modelAndSolution, 0, "configC0", "r4");
    Assignment configD = assertAssignment(modelAndSolution, 0, "configD0", "r3");
    Assignment configE = assertAssignment(modelAndSolution, 0, "configE0", "r2");
    Assignment configF = assertAssignment(modelAndSolution, 0, "configF0", "r5");
    Assignment configG = assertAssignment(modelAndSolution, 0, "configG0", "r6");
    assertComponentRequirement(configA,"beta", configB);
    assertComponentRequirement(configA,"epsilon", configE);
    assertComponentRequirement(configB,"chi", configC);
    assertComponentRequirement(configB,"delta", configD);
    assertComponentRequirement(configE,"phi", configF);
    assertComponentRequirement(configE,"gamma", configG);
    assertEquals("Objective does not fit", 21885.66,
        modelAndSolution.getSecondElement().computeObjective(), 0.01);
  }

  @Test
  public void test_multi_01() throws IOException, Parser.Exception, SolvingException {
    Tuple<Root, Solution> modelAndSolution = loadAndSolve("test_multi_01.txt");
    assertValidSolution(modelAndSolution);
    Assignment config_0i0m0 = assertAssignment(modelAndSolution, 0, "config_0i0m0", "r0");
    Assignment config_1i0m0 = assertAssignment(modelAndSolution, 0, "config_1i0m0", "r0");
    assertComponentRequirement(config_0i0m0,"other", config_1i0m0);
  }

}
